{% extends "aa_forum/base.html" %}

{% load aa_forum_versioned_static %}
{% load static %}
{% load i18n %}

{% block page_title %}{% translate "Forum (Administration)" %}{% endblock %}

{% block aa_forum_header %}
    <br>

    {% include "aa_forum/partials/menu.html" %}
{% endblock %}

{% block aa_forum_body %}
    <script type="application/javascript">
        let aaForumAdminSettings = {}
    </script>

    {% include "aa_forum/partials/administration/categories.html" %}
{% endblock %}

{% block extra_css %}
    {% include "bundles/jquery-ui-css.html" %}

    <link rel="stylesheet" type="text/css" href="{% aa_forum_static 'aa_forum/css/aa-bootstrap-fix.min.css' %}">
    <link rel="stylesheet" type="text/css" href="{% static 'aa_forum/libs/sumoselect/3.0.2-66/sumoselect.min.css' %}">
    <link rel="stylesheet" type="text/css" href="{% aa_forum_static 'aa_forum/css/aa-forum.min.css' %}">
{% endblock %}

{% block extra_javascript %}
    {% include "bundles/jquery-ui-js.html" %}

    <script type="application/javascript" src="{% static 'aa_forum/libs/sumoselect/3.0.2-66/jquery.sumoselect.min.js' %}"></script>
    <script type="application/javascript" src="{% aa_forum_static 'aa_forum/javascript/aa-forum-admin.min.js' %}"></script>
{% endblock %}
